/*
* Copyright 2016 Nu-book Inc.
* Copyright 2016 ZXing authors
*/
// SPDX-License-Identifier: Apache-2.0

#pragma once

#include <array>

namespace ZXing::OneD::Code128 {

extern const std::array<std::array<int, 6>, 107> CODE_PATTERNS;

} // namespace ZXing::OneD::Code128
